/************************************************************************/
/*  Rocket.java								*/
/*									*/
/*  Greenspan Discrete Mechanics for a rocket                   	*/
/*  This records the rocket state and provides methods for initializing	*/
/*  updating same.							*/
/*									*/
/*  Rocket design is based on the Bussard-Jameson ARC/QED propulsion	*/
/*  system, with vehicle parameters adapted from the STS orbiter.	*/
/************************************************************************/

import java.lang.Math;
import java.io.*;
import java.net.*;

public class Rocket
 {
  public int limit = 10,	/* Max iterations per time step 	*/
	      fail,		/* No. of convergence failures  	*/
		 N;             /* No. of Massive Bodies		*/

/**************    Rocket State data	*********************************/

  public double Xr0[],		/* Current position			*/
                Xr1[],		/* Next position			*/
	        Vr0[],		/* Current velocity			*/
	        Vr1[],		/* Next velocity			*/
	        TV[],		/* Thrust vector (length=1)     	*/
	        A0[],		/* Current Acceleration			*/
	        Mr0,		/* Current mass				*/ 
	        Mr1,            /* New mass				*/
	        Pr,		/* Current Propellant			*/
                Thr,            /* Throttle setting			*/
	        dm,		/* change in mass			*/
	        DPr,		/* propellant consumption at 100%  	*/
                stime,		/* system time for X0, V0		*/
	        deltat,		/* current delta-t			*/
	        epsilon,	/* convergence criterion		*/
	        Is,		/* Specific Impulse             	*/
		IsTh = 1600.0,	/* Is*Thr (constant)			*/
		g,		/* Acceleration of gravity      	*/
		fuel_limit,	/* Point of empty tank			*/
		iPr,		/* Intial propellant			*/
                Accel;		/* Acceleration to pass to control	*/

  Object thsync,		/* throttle synchronization		*/
         TVsync;		/* Thrust vector synchronization	*/

                 
/***********************************************************************/


Rocket(double t0)	/* starting time	*/
 {  stime = t0;
    DPr   = .8;
    Xr0   = new double[3];
    Xr1   = new double[3];
    Vr0   = new double[3];
    Vr1   = new double[3];
    TV    = new double[3];
    A0    = new double[3];
    thsync = new Object();  TVsync = new Object();
 }
/***********************************************************************/

/**** Initialize the rocket  ******/

public void initR(Body b, double im, double ipr, double ithr,
		  double gaccel, double spcimp, double deltaP, int Nbods)
	{ Xr0[0] = b.X[0]; Xr0[1] = b.X[1]; Xr0[2] = b.X[2];
	  Vr0[0] = b.V[0]; Vr0[1] = b.V[1]; Vr0[2] = b.V[2];
          Mr0 = im;	/** Initial Mass **/
	  Pr = ipr;	/** initial propellant **/
	  iPr = ipr;	/** for per cent propellant **/
	  Thr = ithr;
	  g = gaccel;
	  IsTh = spcimp;
	  DPr = deltaP;
	  N = Nbods;
	  fuel_limit = im - ipr; 
	}

/************************************************/
/***** Throttle function ************************/
/************************************************/

/*** Computes the change in mass based upon propellant consumption ***/
/*** derived from the throttle setting                             ***/
/*** This engine has consumption proportional to Thr^2.		   ***/
/*** Lowest operating throttle is 20%.				   ***/

void throttle()
	{ synchronized ( thsync )
	   { if ( Thr >= 20.0 )
	      { dm = deltat * DPr * Thr*Thr;
		Is = IsTh/Thr;
	      }
	     else
	      { dm = 0.0;  Is = 0.0; }
	     // System.out.println(" /  "+Thr+"   dm: "+dm+"   Is: "+Is);
	   }
	  
	}


/************* Prepare rocket for integration *************/


void ready(double X0[][],  double Mg[])
  { throttle();
    for (int i = 0;  i < 3;  i++)  Xr1[i] = Xr0[i];
    RocketAccel(X0, X0, Mg);
  }
    


/*******  deltav used for Is * g * log(Mr0/Mr1) * TV ******/
   
double deltav[] = new double[3];



/*******  Update deltav; needed to avoid Netscape restrictions	******/
/*******  on synchronization.					******/

void upDeltav(double cdv)
 { synchronized ( TVsync )
    { for ( int i = 0; i < 3; i++ )  deltav[i] = cdv * TV[i]; }
 }



/************************************************/
/***** One Step function ************************/

public int OneRstep(double X0[][], double X1[][], double Mg[])
  { double mx;  int i, ic;
    double cdv;
   
    if (dm > 0.0)
    	{if (Pr <= 0.0) {DPr = 0.0; throttle();}
       	   else Pr = Pr - dm;};     
    Mr1 = Mr0 - dm;
    cdv = Is * g * Math.log(Mr0/Mr1);
    upDeltav(cdv);
    for ( i = 0; i < 3; i++ )
    { Vr1[i] = Vr0[i] + (deltat * A0[i]) + deltav[i];
      Xr1[i] = Xr0[i] + (0.5 * deltat * (Vr0[i] + Vr1[i]));
    }
    
    Accel = ( cdv / deltat ) / g;   /*  This is passed to RControl  */
	    
    mx = RupdateXV() + 2.0 * epsilon;

    for ( ic = 0; (mx > epsilon) && (ic < limit); ic++ )
     {
	RocketAccel(X0, X1, Mg);
	mx = RupdateXV();
     }
    if ( mx > epsilon ) 
     { fail++;  
       System.out.println("Mass: "+Mr0+"  mx: "+mx+"  ic:"+ic);
     }
    Mr0 = Mr1;
    return ic;
  }

/************************************************/
/***** Set acceleration to zero *****************/

void clearAc()
 {
   for (int i=0; i<3; i++ )
    {  A0[i] = 0.0;  }
 }  

/************************************************/
/***** Rocket Acceleration **********************/


public void RocketAccel(double X0[][], double X1[][], double Mg[])
 { double Dir[],			/* (X0[i]-Xr0) + (X1[i]-Xr1)  */
	  r0, r1,        		/* body i and rocket r	      */
	  fi,				/* multiplier for body i      */
	  fr,				/* multiplier for rocket r    */
	  w,v;				/* miscellaneous 	      */
   int i,u;			        /* counter for loop 	      */
      
   Dir = new double[3];  clearAc();
   
   for ( i = 0; i < N; i++ )
     {
	r0 = 0.0;  r1 = 0.0;
	for ( u = 0; u < 3; u++ )
	{ 
	  w = X0[i][u] - Xr0[u];   r0 += w*w;
	  v = X1[i][u] - Xr1[u];   r1 += v*v;
	  Dir[u] = w + v;
	}
       r0 = Math.sqrt(r0);   r1 = Math.sqrt(r1);  w = r0*r1*(r0 + r1);
       fi = Mg[i]/w;
       for ( u = 0; u < 3; u++ )
       { A0[u] += fi*Dir[u];}
     }
    /* System.out.println("A0: "+A0[0]+"  "+A0[1]+"  "+A0[2]); */
}	
	
/*****************************************************/
/***** Rocket Update *********************************/


double RupdateXV()
  { double mx, w, hdt = 0.5*deltat;   int i;

    mx = 0.0;
    
    for ( i = 0; i < 3; i++ )
     {  w = Vr0[i] + deltat * A0[i] + deltav[i];
	mx = Math.max(mx, Math.abs(Vr1[i] - w));
	Vr1[i] = w;
	w = Xr0[i] + hdt*(w + Vr0[i]);
	mx = Math.max(mx, Math.abs(Xr1[i] - w));
	Xr1[i] = w;
      }
    return mx;
  }
       
}
   
 

